/*
 * Copyright (C) 2008, 2009 Andrew Sichevoi.
 *
 * This file is part of Conler (http://thekondor.net/conler).
 *
 * Conler is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Conler is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Conler. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h> // required by escape_str()
#include <glib.h>

#ifdef MAEMO
  #include <gconf/gconf.h>
  #include <gconf/gconf-client.h>
#endif

#include "common.h"
#include "const.h"
#include "types.h"

#ifndef RUN_WRAPPER
  #define RUN_WRAPPER   "/usr/bin/conler-run-wrapper.sh"
#endif

static gboolean simulate = FALSE;   /* Simulate mode */
static gchar* config_path = NULL;   /* Alternative config */
static gchar* ap_alt_id = NULL;     /* Alternative AP id */

static GOptionEntry entries[] =
{
  {"config", 'c', 0, G_OPTION_ARG_STRING, &config_path,
    "Path to an alternative configuration file, default: " CONFIG_PATH, "cfg"},
  {"simulate", 's', 0, G_OPTION_ARG_NONE, &simulate,
    "Don't execute commands, only print", NULL},
  {"cid", 'i', 0, G_OPTION_ARG_STRING, &ap_alt_id,
    "Run commands for AP with id = P", "P"},
  { NULL }
};

/* escape_str() -- modified glib's g_strescape(),
   Changes:
    - 'exceptions' removed -- because there is no need in them
    - escaping for single quote added
    - more complex double-quote escaping because of
      process spawn requirements
    - tiny refactoring
*/
static gchar* escape_str(const gchar* source, gboolean escape_unicode)
{
#define ESC(ch, repl) *ch++ = '\\'; *ch++ = repl; break
    const guchar *p;
    gchar *dest;
    gchar *q;
    gboolean odd_quote = TRUE;

    g_return_val_if_fail (source != NULL, NULL);

    p = (guchar *) source;
    /* Each source byte needs maximally four destination chars (\777) */
    q = dest = g_malloc (strlen (source) * 4 + 1);

    while (*p)
    {
	switch (*p)
	{
	   case '\b': ESC(q, 'b');
	   case '\f': ESC(q, 'f');
	   case '\n': ESC(q, 'n');
	   case '\r': ESC(q, 'r');
	   case '\t': ESC(q, 't');
	   case '\\': ESC(q, '\\');
           case '"' : odd_quote = !odd_quote;
	              /* NOTE: Double quotes must be balanced */
	              if (!odd_quote) {
		          *q++ = '\'';
			  *q++ = '\\';
			  *q++ = '"';
	              }
	              else
		      {
			  *q++ = '\\';
			  *q++ = '"';
			  *q++ = '\'';
		      }
		      break;

	   case '\'': ESC(q, '\'');
	   default  :
	             if (escape_unicode) {
	                 if ((*p < ' ') || (*p >= 0177)) {
		            *q++ = '\\';
		            *q++ = '0' + (((*p) >> 6) & 07);
		            *q++ = '0' + (((*p) >> 3) & 07);
		            *q++ = '0' + ((*p) & 07);
		         }
		         else
		            *q++ = *p;
	             }
	             else
	             {
		         *q++ = *p;
	             }

		     break;
	}

	p++;
    }

    *q = 0;
    return dest;
}


static gchar* get_connection_name(const gchar* icd_conn_id)
{
    if (!icd_conn_id)
        return NULL;

    if (DIABLO != get_maemo_version())
        return g_strdup(icd_conn_id);

#ifdef MAEMO
    GConfClient* gconf_client;
    GError* error = NULL;
    gchar* key;
    gchar* conn_name;

    g_type_init();
    gconf_client = gconf_client_get_default();
    if (!gconf_client) {
        g_warning("No GConf client is available!\n");
        return NULL;
    }

    key = g_strdup_printf("/system/osso/connectivity/IAP/%s/name", icd_conn_id);
    conn_name = gconf_client_get_string(gconf_client, key, &error);
    if (error) {
        g_warning("AP name obtaining error: %s\n", error->message);
        g_error_free(error);
    }

    g_free(key);
    g_object_unref(gconf_client);

    return conn_name;
#else
    return g_strdup(icd_conn_id);
#endif
}

static AccessPoint_s* find_access_point(GSList* access_points,
                                        const gchar* const name)
{
    if (!access_points || !name)
        return NULL;

    for (GSList* iter = access_points; iter; iter = g_slist_next(iter))
        if (!g_ascii_strcasecmp(AP_LIST_CAST(iter)->name, name))
            return AP_LIST_CAST(iter);

    return NULL;
}

static void execute_command(const Command_s* cmd, gpointer simulation_mode)
{
  gchar* command = NULL;

  if (cmd->disabled) {
      g_print("'%s' is disabled, skipping\n", cmd->command);
      return;
  }

  if (!cmd->superuser_env) {
      // TODO: re-design process running in non-root environment
      gchar* escaped_command = escape_str(cmd->command, FALSE);
      command = g_strdup_printf(RUN_WRAPPER " \"%s\"", escaped_command);
      g_free(escaped_command);
  }
  else
      command = g_strdup(cmd->command);

  if (*((gboolean *)simulation_mode))
      // TODO: because unicode symbols are not escaped above
      //       they will be represented here as '?'.
      //       Solution: to keep unescaped copy of the command.
      g_print("[%s]: %s\n", cmd->sync_exec ? "fg" : "bg", command);
  else
  {
      GError* error = NULL;
      gint exit_status;

      if (cmd->sync_exec)
	  g_spawn_command_line_sync(command, NULL, NULL, &exit_status, &error);
      else
	  g_spawn_command_line_async(command, &error);
      // TODO: add support of running in the background
      if (error) {
	  g_warning("Command execution error: %s (rc = %d)", error->message,
		                                             exit_status);
	  g_error_free(error);
      }
  }

  g_free(command);
}

/* Check functions */
static void check_connection_type()
{
  const gchar* icd_type = g_getenv("ICD_CONNECTION_TYPE");

  if (icd_type && !g_ascii_strcasecmp(WLAN_BEARER_TYPE, icd_type))
    return;

  g_print("Error: no WiFi connection is available\n");
  exit(1);
}

static void check_config(const gchar* config)
{
  if (g_file_test(config, G_FILE_TEST_EXISTS))
    return;

  g_print("Error: Configuration file '%s' isn't available\n", config);
  exit(1);
}

static void process_config(const gchar* config_path, ConlerCfg_s* cfg)
{
  if (CFG_R_OK == read_cfg(config_path, cfg))
    return;

  g_print("Error: Configuration file can't be parsed\n");
  g_print("Please check '%s' for validness\n", config_path);
  exit(1);
}


/* Here we go */
int main(int argc, char** argv)
{
  ConlerCfg_s cfg;
  GError* error = NULL;
  GOptionContext* context;

  g_set_application_name(PROGRAM_FULL_ID);
  g_print("%s v%s, %s\n\n", PROGRAM_ID, PROGRAM_VERSION, COPYRIGHT);

  context = g_option_context_new("- run AP-specific commands");
  g_option_context_add_main_entries(context, entries, NULL);

  if (!g_option_context_parse(context, &argc, &argv, &error)) {
    g_print("Error: option parsing failed: %s\n", error->message);
    g_print("Use '--help' to see the list of available keys\n");
    g_option_context_free(context);
    exit(1);
  }

 if (!simulate)
    check_connection_type();

  /* Obtain AP */
  gchar* ap_id = ap_alt_id;
  if (!ap_id)
    ap_id = get_connection_name(g_getenv("ICD_CONNECTION_ID"));

  if (!ap_id) {
    g_print("Error: Can't obtain AP name\n");
    exit(1);
  }

  /* Obtain path to the configuration file */
  gchar* config = config_path;
  if (!config_path)
    config = CONFIG_PATH;
  check_config(config);

  process_config(config, &cfg);

  /* Execute AP-specific commands */
  AccessPoint_s* ap;
  ap = find_access_point(cfg.access_points, ap_id);
  if (!ap) {
    g_print("Info: Nothing to do for '%s'\n", ap_id);
    exit(1);
  }

  if (!ap->disabled)
    g_slist_foreach(ap->commands, (GFunc)execute_command, (gpointer)&simulate);
  
  /* Saying thanks to the used memory and releasing it */
  if (config_path)
    g_free(config_path);

  g_free(ap_id);
//  if (ap_alt_id)
//    g_free(ap_alt_id);

//  if (!ap_alt_id && ap_id)
//    g_free(ap_id);
//  else if (ap_alt_id)
//    g_free(ap_alt_id);


  g_option_context_free(context);
  free_cfg(&cfg);

  return 0;
}

